#!/usr/bin/env node
"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ts2md = exports.Ts2Md = void 0;
/* eslint-disable @typescript-eslint/no-unused-vars */
const fs = __importStar(require("fs"));
const path = __importStar(require("path"));
const typescript_1 = __importDefault(require("typescript"));
const mdMerge_1 = require("./mdMerge");
function getJsDoc(node) {
    const jsDoc = (node['jsDoc']) ? node['jsDoc'] : [];
    const tags = [];
    for (const doc of jsDoc) {
        for (const tag of (doc['tags'] ? doc['tags'] : [])) {
            tags.push(tag);
        }
    }
    return { jsDoc, tags };
}
class DocItem {
    constructor(item, name, sf) {
        this.item = item;
        this.name = name;
        this.sf = sf;
        const r = getJsDoc(item);
        this.jsDoc = r.jsDoc;
        this.jsDocTags = r.tags;
        this.isPrivate = r.tags.some(t => t.kind === typescript_1.default.SyntaxKind.JSDocPrivateTag);
    }
}
class DocBase {
    constructor(ts2Md, label, labelPlural) {
        this.ts2Md = ts2Md;
        this.label = label;
        this.labelPlural = labelPlural;
        this.docItems = [];
    }
    tryAddItem(s, sf) {
        const items = this.filterItem(s);
        for (const item of items) {
            const docItem = new DocItem(item, this.getName(item, sf), sf);
            if (!docItem.isPrivate)
                this.docItems.push(docItem);
        }
    }
    toMarkDown(docItem) {
        let md = `${this.ts2Md.headingLevelMd(3)} ${this.label}: ${docItem.name}\n\n`;
        if (docItem.jsDoc.some(d => d.kind === typescript_1.default.SyntaxKind.JSDoc)) {
            md += `${this.ts2Md.headingLevelMd(4)} Description\n\n`;
            for (const d of docItem.jsDoc) {
                if (d['comment'])
                    md += `${d['comment']}\n\n`;
            }
        }
        md += '```ts\n';
        md += this.toMarkDownTs(docItem);
        md += '\n```\n\n';
        md += this.toMarkDownDetails(docItem);
        return md;
    }
    toMarkDownTs(docItem) {
        return this.ts2Md.printer.printNode(typescript_1.default.EmitHint.Unspecified, docItem.item, docItem.sf);
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    toMarkDownDetails(docItem) {
        return '';
    }
    toMarkDownRefLink(docItem) {
        return `[${docItem.name}](#${this.label.toLowerCase()}-${docItem.name.toLowerCase()})`;
    }
    isExportedDeclaration(item) {
        const modFlags = typescript_1.default.getCombinedModifierFlags(item);
        const hasExportFlag = !!(modFlags & typescript_1.default.ModifierFlags.Export);
        return hasExportFlag;
    }
}
class DocVariable extends DocBase {
    constructor(ts2Md) { super(ts2Md, 'Variable', 'Variables'); }
    getName(item, sf) {
        return item.name.getText(sf);
    }
    filterItem(item) {
        const items = [];
        if (typescript_1.default.isVariableStatement(item) && item.modifiers && item.modifiers.some(m => m.kind === typescript_1.default.SyntaxKind.ExportKeyword)) {
            for (const v of item.declarationList.declarations) {
                items.push(v);
            }
        }
        return items;
    }
}
class DocType extends DocBase {
    constructor(ts2Md) { super(ts2Md, 'Type', 'Types'); }
    getName(item) {
        return item.name.text;
    }
    filterItem(item) {
        if (typescript_1.default.isTypeAliasDeclaration(item) && this.isExportedDeclaration(item))
            return [item];
        return [];
    }
}
class DocInterface extends DocBase {
    constructor(ts2Md) { super(ts2Md, 'Interface', 'Interfaces'); }
    getName(item) {
        return item.name.text;
    }
    filterItem(item) {
        if (typescript_1.default.isInterfaceDeclaration(item) && this.isExportedDeclaration(item))
            return [item];
        return [];
    }
    toMarkDownDetails(docItem) {
        let md = '';
        let detailsStartAdded = false;
        const addDetailsStart = () => {
            if (!detailsStartAdded) {
                detailsStartAdded = true;
                md += `<details>\n\n<summary>${this.label} ${docItem.name} Member Details</summary>\n\n`;
            }
        };
        const addDetailsEnd = () => {
            if (detailsStartAdded) {
                md += `</details>\n\n`;
            }
        };
        for (const m of docItem.item.members) {
            const ps = m;
            const propName = ps.name.escapedText;
            const r = getJsDoc(m);
            if (r.jsDoc.length > 0) {
                addDetailsStart();
                md += `${this.ts2Md.headingLevelMd(5)} ${propName}\n\n`;
                for (const d of r.jsDoc) {
                    if (d['comment']) {
                        md += `${d['comment']}\n\n`;
                    }
                }
            }
        }
        addDetailsEnd();
        return md;
    }
}
class DocFunction extends DocBase {
    constructor(ts2Md) { super(ts2Md, 'Function', 'Functions'); }
    getName(item) {
        var _a;
        return ((_a = item.name) === null || _a === void 0 ? void 0 : _a.text) || '';
    }
    filterItem(item) {
        if (typescript_1.default.isFunctionDeclaration(item) && this.isExportedDeclaration(item))
            return [item];
        return [];
    }
    toMarkDownTs(docItem) {
        const n = docItem.item;
        const sf = docItem.sf;
        const printer = this.ts2Md.printer;
        let mdts = printer.printNode(typescript_1.default.EmitHint.Unspecified, n, sf);
        if (n.body && !docItem.jsDocTags.some(t => t.kind === typescript_1.default.SyntaxKind.JSDocTag
            && t.tagName.escapedText === 'publicbody')) {
            const bodyts = printer.printNode(typescript_1.default.EmitHint.Unspecified, n.body, sf);
            mdts = mdts.slice(0, mdts.length - bodyts.length);
        }
        return mdts;
    }
    toMarkDownDetails(docItem) {
        let md = '';
        const returnsTags = docItem.jsDocTags.filter(t => t.kind === typescript_1.default.SyntaxKind.JSDocReturnTag)
            .map(t => t)
            .filter(t => !!t.comment);
        if (returnsTags.length > 0) {
            md += `${this.ts2Md.headingLevelMd(4)} Returns\n\n`;
            for (const t of returnsTags) {
                const rt = t;
                md += `${rt.comment}\n\n`;
            }
        }
        const paramTags = docItem.jsDocTags.filter(t => t.kind === typescript_1.default.SyntaxKind.JSDocParameterTag).map(t => t).filter(t => t.comment);
        if (paramTags.length > 0) {
            md += `<details>\n\n<summary>${this.label} ${docItem.name} Argument Details</summary>\n\n`;
            for (const tag of paramTags) {
                const name = tag.name.getText(docItem.sf);
                md += `${this.ts2Md.headingLevelMd(5)} ${name}\n\n${tag.comment}`;
            }
            md += `</details>\n\n`;
        }
        return md;
    }
}
class DocClass extends DocBase {
    constructor(ts2Md) {
        super(ts2Md, 'Class', 'Classes');
        this.details = {};
    }
    getName(item) {
        var _a;
        return ((_a = item.name) === null || _a === void 0 ? void 0 : _a.text) || '';
    }
    filterItem(item) {
        if (typescript_1.default.isClassDeclaration(item) && this.isExportedDeclaration(item))
            return [item];
        return [];
    }
    findTs(findInTs, targetTs) {
        let pos = findInTs.indexOf(targetTs);
        if (pos === -1) {
            // Try outdenting once
            targetTs = targetTs.replace(/\n/g, '\n    ');
            pos = findInTs.indexOf(targetTs);
            if (pos === -1) {
                // set and get accessor bodies sometimes are inlined in full generated class typescript
                // remove all indenting
                targetTs = targetTs.replace(/\n */g, ' ');
                pos = findInTs.indexOf(targetTs);
            }
        }
        return { pos, len: targetTs.length };
    }
    removeTs(fromTs, removeTs, withSemi) {
        const r = this.findTs(fromTs, removeTs);
        // See if we would leave a dangling semicolon behind
        if (r.pos > -1 && withSemi && r.pos + r.len + 1 < fromTs.length && fromTs[r.pos + r.len] === ';' && fromTs[r.pos + r.len + 1] === '\n') {
            r.len += 2;
            // and remove leading spaces
            while (r.pos > 0 && fromTs[r.pos - 1] === ' ') {
                r.pos--;
                r.len++;
            }
        }
        if (r.pos > -1) {
            fromTs = fromTs.slice(0, r.pos) + fromTs.slice(r.pos + r.len);
            if (fromTs[r.pos] === '\n') {
                let pos2 = r.pos - 1;
                while (pos2 > 0 && fromTs[pos2] === ' ')
                    pos2--;
                if (fromTs[pos2] === '\n') {
                    // and remove blank line left after original removal
                    fromTs = fromTs.slice(0, pos2) + fromTs.slice(r.pos);
                }
            }
        }
        return fromTs;
    }
    toMarkDownTs(docItem) {
        var _a, _b;
        const n = docItem.item;
        const sf = docItem.sf;
        const printer = this.ts2Md.printer;
        this.details = {};
        // class definition typescript including all members and function bodies...
        let mdts = printer.printNode(typescript_1.default.EmitHint.Unspecified, n, sf);
        for (const ce of n.members) {
            const r = getJsDoc(ce);
            let isPrivate = false;
            if ((ce['modifiers'] && ce['modifiers'].some(m => m.kind === typescript_1.default.SyntaxKind.PrivateKeyword)) ||
                r.tags.some(t => typescript_1.default.isJSDocPrivateTag(t))) {
                isPrivate = true;
                // Remove entire member
                const memberts = printer.printNode(typescript_1.default.EmitHint.Unspecified, ce, sf);
                mdts = this.removeTs(mdts, memberts, true);
            }
            else if (ce['body'] && !r.tags.some(t => t.tagName.escapedText === 'publicbody')) {
                // Remove the body from documentation typescript
                const bodyts = printer.printNode(typescript_1.default.EmitHint.Unspecified, ce['body'], sf);
                mdts = this.removeTs(mdts, bodyts);
            }
            if (!isPrivate && r.jsDoc.length > 0) {
                // Collect member details
                let name = ((_a = ce.name) === null || _a === void 0 ? void 0 : _a.getText(docItem.sf)) || '';
                if (ce.kind === typescript_1.default.SyntaxKind.Constructor) {
                    name = 'constructor';
                    // Default value of {}['constructor'] is an object...
                    if (typeof this.details[name] !== 'string')
                        this.details[name] = '';
                }
                let memberDetails = (_b = this.details[name]) !== null && _b !== void 0 ? _b : "";
                for (const d of r.jsDoc) {
                    if (d['comment']) {
                        memberDetails += `${d['comment']}\n\n`;
                    }
                }
                this.details[name] = memberDetails;
            }
        }
        return mdts;
    }
    toMarkDownDetails(docItem) {
        const keys = Object.keys(this.details);
        if (keys.length < 1)
            return '';
        let md = `<details>\n\n<summary>${this.label} ${docItem.name} Member Details</summary>\n\n`;
        keys.sort();
        for (const key of keys) {
            md += `${this.ts2Md.headingLevelMd(5)} ${key}\n\n${this.details[key]}`;
        }
        md += `</details>\n\n`;
        return md;
    }
}
/**
 * Uses the Typescript compiler to parse source tree given a top level source file such as `index.ts`.
 *
 * Extract the exported API interfaces, classes, types, functions and variables.
 *
 * Generate GitHub friendly MarkDown documentation for the extracted API leveraging TypeScript type information
 * and merging JSDoc style documentation comments.
 *
 * The following JSDoc tags are supported:
 *
 *    `@publicbody` Applied to a class method or function. Adds the function body to the documentation with embedded comments removed.
 *
 *    `@private` Applied to an exported or publicly accessible member keeps it out of documentation.
 */
class Ts2Md {
    /**
     * Construct a new instance configured for `run` method to be called next.
     *
     * @param options Must be provided. inputFilename defaults to `./src/index.ts`
     */
    constructor(options) {
        this.options = options;
        this.sourceFiles = [];
        this.docs = [];
        options.inputFilename || (options.inputFilename = './src/index.ts');
        this.filePath = path.resolve(options.inputFilename);
        this.fileName = path.parse(this.filePath).name;
        const compOpts = {
            allowJs: false,
            target: typescript_1.default.ScriptTarget.Latest
        };
        const progOpts = {
            rootNames: [this.filePath],
            options: compOpts
        };
        this.program = typescript_1.default.createProgram(progOpts);
        const printOpts = {
            newLine: typescript_1.default.NewLineKind.LineFeed,
            removeComments: true,
            omitTrailingSemicolon: true
        };
        const printHandlers = {};
        this.printer = typescript_1.default.createPrinter(printOpts, printHandlers);
    }
    /**
     * Generates the documentation markdown and write's it to output file
     * and/or merges it to README.md
     */
    run() {
        this.sourceFiles = this.parseSourceFiles(this.program);
        this.docs = this.extractDocs(this.sourceFiles);
        this.markDown = this.generateMarkDown(this.docs);
        if (this.markDown) {
            if (this.options.outputFilename)
                this.outputPath = this.writeMarkDownToOuput(this.markDown, this.options.outputFilename);
            if (this.options.readmeMerge)
                this.readmeMerge(this.markDown);
        }
    }
    /**
     * @private
     */
    headingLevelMd(relativeLevel) {
        relativeLevel += this.options.firstHeadingLevel - 1;
        if (this.options.noTitle)
            relativeLevel--;
        return '#'.repeat(relativeLevel + this.options.firstHeadingLevel - 1);
    }
    parseSourceFiles(program) {
        return program.getSourceFiles().filter(sf => sf.fileName.indexOf('node_modules') === -1);
    }
    extractDocs(sourceFiles) {
        let docs = [
            new DocInterface(this),
            new DocClass(this),
            new DocFunction(this),
            new DocType(this),
            new DocVariable(this)
        ];
        for (const sf of sourceFiles) {
            console.log(`SourceFile ${sf.fileName}`);
            typescript_1.default.forEachChild(sf, node => {
                if (typescript_1.default.isStatement(node)) {
                    for (const doc of docs) {
                        doc.tryAddItem(node, sf);
                    }
                }
            });
        }
        // Eliminate empty doc categories
        docs = docs.filter(d => d.docItems.length > 0);
        return docs;
    }
    generateMarkDown(docs) {
        let md = '';
        if (!this.options.noTitle) {
            md += `${this.headingLevelMd(1)} API\n\n`;
        }
        else {
            md += '\n';
        }
        let linksMd = `Links: [API](#api)`;
        for (const doc of docs) {
            linksMd += `, [${doc.labelPlural}](#${doc.labelPlural.toLowerCase()})`;
        }
        linksMd += '\n\n';
        md += linksMd;
        for (const doc of this.docs) {
            md += `${this.headingLevelMd(2)} ${doc.labelPlural}\n\n`;
            if (doc.docItems.length > 1) {
                md += this.generateDocItemLinksTable(doc);
                md += `\n${linksMd}---\n\n`;
            }
            for (const docItem of doc.docItems) {
                md += doc.toMarkDown(docItem);
                md += linksMd;
                md += '---\n';
            }
        }
        return md;
    }
    generateDocItemLinksTable(doc) {
        let md = '';
        const nameLinks = doc.docItems.map(d => ({ name: d.name, linkMd: doc.toMarkDownRefLink(d) }));
        nameLinks.sort((a, b) => a.name < b.name ? -1 : a.name === b.name ? 0 : 1);
        let cols = 1, rows = nameLinks.length;
        while (cols < 4 && rows > 12) {
            cols++;
            rows = Math.ceil(nameLinks.length / cols);
        }
        for (let col = 1; col <= cols; col++) {
            md += '| ';
        }
        md += '|\n';
        for (let col = 1; col <= cols; col++) {
            md += '| --- ';
        }
        md += '|\n';
        for (let row = 1; row <= rows; row++) {
            for (let col = 1; col <= cols; col++) {
                const i = (col - 1) * rows + row - 1;
                const d = i < nameLinks.length ? nameLinks[i].linkMd : '';
                if (col === 1)
                    md += '|';
                md += ` ${d} |`;
            }
            md += '\n';
        }
        return md;
    }
    writeMarkDownToOuput(markDown, outputFilename) {
        const outputPath = path.resolve(outputFilename);
        if (this.options.outputReplace) {
            try {
                fs.unlinkSync(outputPath);
            }
            catch ( /* */_a) { /* */ }
        }
        fs.writeFileSync(outputPath, markDown);
        return outputPath;
    }
    readmeMerge(markDown) {
        (0, mdMerge_1.mdMerge)(markDown);
    }
}
exports.Ts2Md = Ts2Md;
/**
 * Generate Typescript documentation and merge into README.md
 *
 * Attempts to validate options, constructs an instance of Ts2Md with those options, and runs the generation method.
 *
 * 1. Function argument is used if provided.
 *
 * 2. Looks for `./ts2md.json`
 *
 * 3. Default options.
 *
 * Default options are:
 *
 * ```json
 * {
 *   "inputFilename": "./src/index.ts",
 *   "outputFilename": "./apiDoc.md",
 *   "firstHeadingLevel": 2,
 *   "noTitle": true,
 *   "outputReplace": true,
 *   "readmeMerge": true
 * }
 * ```
 * @publicbody
 */
function ts2md(options) {
    if (!options) {
        try {
            const configPath = path.resolve('./ts2md.json');
            const json = fs.readFileSync(configPath, { encoding: 'utf8' });
            options = JSON.parse(json);
        }
        catch ( /* */_a) { /* */ }
    }
    options || (options = {
        inputFilename: './src/index.ts',
        outputFilename: '',
        firstHeadingLevel: 2,
        noTitle: true,
        outputReplace: true,
        readmeMerge: true
    });
    new Ts2Md(options).run();
}
exports.ts2md = ts2md;
console.log('running ts2md()');
ts2md();
//# sourceMappingURL=ts2md.js.map